/**
 * Copyright (c) 2014 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather;

import java.util.ArrayList;
import java.util.StringTokenizer;


/**
 * A class to calculate the major and minor time intervals based on
 * the number of minutes being shown by the plot and the configuration.
 * Mostly this class simply parses a configuration string on startup,
 * and then later returns the appropriate interval using the parsed data.
 */
public class TimeIntervals
{
	/**
	 * A table of the time interval information objects.
	 * There is no particular order to the elements.
	 */
	private ArrayList<IntervalInfo> mIntervalTable;


	/**
	 * Format of the configuration string:
	 *	spec;spec...
	 * where each spec is:
	 * 	plotTime:majorInterval:minorInterval
	 * each of plotTime, majorInterval, and minorInterval is a number
	 * of minutes, specified by a number followed by one of 'm', 'h', 'd',
	 * indicating minutes, hours, or days.  For example:
	 * 	"5d:1h:30m;30d:1d:12h"
	 * The semicolons can alternatively be spaces, and the suffix
	 * characters can be capitalised.
	 * @param configString
	 */
	public TimeIntervals(String configString)
	{
		mIntervalTable = new ArrayList<IntervalInfo>();

		//
		// Break the configuration string into individual specifications
		// using spaces or semi-colons and then parse each spec.
		//
		StringTokenizer tokenizer = new StringTokenizer(configString, "; ");

		boolean isFailed = false;

		while (tokenizer.hasMoreTokens())
		{
			String spec = tokenizer.nextToken();

			if (spec.isEmpty())
				continue;

			if (!parseSpec(spec))
				isFailed = true;
		}

		if (isFailed)
			Utils.logError("Invalid time interval specification: " + configString);
	}


	/**
	 * Parse one spec string token.  This is of the form:
	 *	plotTime:majorInterval:minorInterval
	 * Returns true if the time spec is reasonable with the parsed
	 * result stored for later use.
	 * @param spec
	 * @return
	 */
	private boolean parseSpec(String spec)
	{
		//
		// Extract the plot time string.
		//
		int colonIndex = spec.indexOf(':');

		if (colonIndex <= 0)
			return false;

		String plotString = spec.substring(0, colonIndex);
		spec = spec.substring(colonIndex + 1);

		//
		// Extract the major interval time string.
		//
		colonIndex = spec.indexOf(':');

		if (colonIndex <= 0)
			return false;

		String majorString = spec.substring(0, colonIndex);
		spec = spec.substring(colonIndex + 1);

		//
		// Make sure there are no more colons.
		//
		if (spec.indexOf(':') >= 0)
			return false;

		String minorString = spec;

		//
		// Now parse the time values and fail if they are bad.
		//
		int plotTime = parseTime(plotString);
		int majorTime = parseTime(majorString);
		int minorTime = parseTime(minorString);

		if ((plotTime <= 0) || (majorTime <= 0) || (minorTime <= 0))
			return false;

		//
		// Make sure the values seem consistent.
		// This means that they are ordered correctly, that they divide
		// each other, and that the minor interval is a multiple of 30.
		//
		if ((majorTime > plotTime) || (minorTime > majorTime))
			return false;

		if ((plotTime % majorTime) != 0)
			return false;

		if ((majorTime % minorTime) != 0)
			return false;

		if ((minorTime % 30) != 0)
			return false;

		//
		// The times are reasonable so store the values into the table.
		//
		IntervalInfo info = new IntervalInfo(plotTime, majorTime, minorTime);

		mIntervalTable.add(info);

		return true;
	}


	/**
	 * Parse one time string value to return a number of minutes.
	 * This is a number followed by one of the characters 'm', 'h', or 'd' in either case.
	 * For example, "12h".  Returns the number of minutes or -1 on an error.
	 */
	private int parseTime(String timeString)
	{
		if (timeString.length() < 2)
			return -1;

		char suffix = timeString.charAt(timeString.length() - 1);
		timeString = timeString.substring(0, timeString.length() - 1);

		int multiplier;

		switch (suffix)
		{
			case 'm': case 'M': multiplier = 1; break;
			case 'h': case 'H': multiplier = 60; break;
			case 'd': case 'D': multiplier = 60 * 24; break;
			default: return -1;
		}

		int value;

		try
		{
			value = Integer.parseInt(timeString);
		}
		catch (Exception e)
		{
			return -1;
		}

		if (value <= 0)
			return -1;

		return value * multiplier;
	}


	/**
	 * Find the appropriate time interval information object for
	 * the specified plot time in minutes.  This is the interval whose
	 * plot time is as large as possible but still at least as small as
	 * the specified plot time.  Returns null if one cannot be found.
	 * @param plotTIme
	 * @return
	 */
	public IntervalInfo getTimeInterval(int plotTime)
	{
		IntervalInfo bestInfo = null;

		for (IntervalInfo info: mIntervalTable)
		{
			if (info.mPlotTime > plotTime)
				continue;

			if ((bestInfo == null) ||
				(info.mPlotTime > bestInfo.mPlotTime))
			{
				bestInfo = info;
			}
		}
	
		return bestInfo;
	}


	/**
	 * An immutable class to hold time interval information.
	 * All of the values are in minutes.
	 */
	public static class IntervalInfo
	{
		public final int mPlotTime;
		public final int mMajorTime;
		public final int mMinorTime;


		public IntervalInfo(int plotTime, int majorTime, int minorTime)
		{
			mPlotTime = plotTime;
			mMajorTime = majorTime;
			mMinorTime = minorTime;
		}
	}
}
